@@ -47,6 +47,8 @@
 -- >
 -- > main = withSocketsDo
 -- >      $ simpleHttp "http://www.haskell.org/" >>= L.putStr
+-- >
+-- > Cookies are implemented according to RFC 6265.
 module Network.HTTP.Conduit
     ( -- * Perform a request
       simpleHttp
@@ -68,6 +70,7 @@ module Network.HTTP.Conduit
     , requestHeaders
     , requestBody
     , proxy
+    , socksProxy
     , rawBody
     , decompress
     , redirectCount
@@ -83,11 +86,22 @@ module Network.HTTP.Conduit
     , managerCheckCerts
       -- *** Defaults
     , defaultCheckCerts
+      -- * Cookies
+    , Cookie(..)
+    , CookieJar
+    , createCookieJar
+    , destroyCookieJar
+    , updateCookieJar
+    , receiveSetCookie
+    , insertCookiesIntoRequest
+    , computeCookieString
+    , evictExpiredCookies
       -- * Utility functions
     , parseUrl
     , applyBasicAuth
     , addProxy
     , lbsResponse
+    , getRedirectedRequest
       -- * Decompression predicates
     , alwaysDecompress
     , browserDecompress
@@ -103,7 +117,6 @@ module Network.HTTP.Conduit
 
 import qualified Data.ByteString as S
 import qualified Data.ByteString.Lazy as L
-import qualified Data.ByteString.Char8 as S8
 
 import qualified Network.HTTP.Types as W
 import Data.Default (def)
@@ -117,10 +130,13 @@ import Data.Conduit.Blaze (builderToByteString)
 import Control.Monad.Trans.Resource (ResourceT, ResourceIO)
 import Control.Exception.Lifted (try, SomeException)
 
+import Data.Time.Clock
+
 import Network.HTTP.Conduit.Request
 import Network.HTTP.Conduit.Response
 import Network.HTTP.Conduit.Manager
 import Network.HTTP.Conduit.ConnInfo
+import Network.HTTP.Conduit.Cookies
 
 -- | The most low-level function for initiating an HTTP request.
 --
@@ -150,50 +166,22 @@ http req0 manager = do
     res@(Response status hs body) <-
         if redirectCount req0 == 0
             then httpRaw req0 manager
-            else go (redirectCount req0) req0
+            else go (redirectCount req0) req0 def
     case checkStatus req0 status hs of
         Nothing -> return res
         Just exc -> do
-            body' <- C.prepareSource body
-            C.sourceClose body'
+            C.sourceClose body
             liftBase $ throwIO exc
   where
-    go 0 _ = liftBase $ throwIO TooManyRedirects
-    go count req = do
-        res@(Response (W.Status code _) hs _) <- httpRaw req manager
-        case (300 <= code && code < 400, lookup "location" hs) of
-            (True, Just l'') -> do
-                -- Prepend scheme, host and port if missing
-                let l' =
-                        case S8.uncons l'' of
-                            Just ('/', _) -> concat
-                                [ "http"
-                                , if secure req then "s" else ""
-                                , "://"
-                                , S8.unpack $ host req
-                                , ":"
-                                , show $ port req
-                                , S8.unpack l''
-                                ]
-                            _ -> S8.unpack l''
-                l <- liftBase $ parseUrl l'
-                let req' = req
-                        { host = host l
-                        , port = port l
-                        , secure = secure l
-                        , path = path l
-                        , queryString = queryString l
-                        , method =
-                            -- According to the spec, this should *only* be for
-                            -- status code 303. However, almost all clients
-                            -- mistakenly implement it for 302 as well. So we
-                            -- have to be wrong like everyone else...
-                            if code == 302 || code == 303
-                                then "GET"
-                                else method l
-                        }
-                go (count - 1) req'
-            _ -> return res
+    go 0 _ _ = liftBase $ throwIO TooManyRedirects
+    go count req'' cookie_jar'' = do
+        now <- liftIO $ getCurrentTime
+        let (req', cookie_jar') = insertCookiesIntoRequest req'' (evictExpiredCookies cookie_jar'' now) now
+        res' <- httpRaw req' manager
+        let (cookie_jar, res) = updateCookieJar res' req' now cookie_jar'
+        case getRedirectedRequest req' (responseHeaders res) (W.statusCode (statusCode res)) of
+            Just req -> go (count - 1) req cookie_jar
+            Nothing -> return res
 
 -- | Get a 'Response' without any redirect following.
 httpRaw
