@@ -28,9 +28,11 @@ import Data.ByteString (ByteString)
 import qualified Data.ByteString as S
 import qualified Data.ByteString.Lazy as L
 
+import Network (PortID(..))
 import Network.Socket (Socket, sClose)
 import Network.Socket.ByteString (recv, sendAll)
 import qualified Network.Socket as NS
+import Network.Socks5 (socksConnectWith, SocksConf)
 
 import Network.TLS
 import Network.TLS.Extra (ciphersuite_all)
@@ -54,20 +56,23 @@ data ConnInfo = ConnInfo
     }
 
 connSink :: C.ResourceIO m => ConnInfo -> C.Sink ByteString m ()
-connSink ConnInfo { connWrite = write } = C.Sink $ return $ C.SinkData
-    { C.sinkPush = \bss -> liftBase (write bss) >> return C.Processing
-    , C.sinkClose = return ()
-    }
+connSink ConnInfo { connWrite = write } =
+    C.SinkData push close
+  where
+    push bss = liftBase (write bss) >> return (C.Processing push close)
+    close = return ()
 
 connSource :: C.ResourceIO m => ConnInfo -> C.Source m ByteString
-connSource ConnInfo { connRead = read' } = C.Source $ return $ C.PreparedSource
-    { C.sourcePull = do
+connSource ConnInfo { connRead = read' } =
+    src
+  where
+    src = C.Source pull close
+    pull = do
         bs <- liftBase read'
         if S.null bs
             then return C.Closed
-            else return $ C.Open bs
-    , C.sourceClose = return ()
-    }
+            else return $ C.Open src bs
+    close = return ()
 
 #if DEBUG
 allOpenSockets :: I.IORef (Int, IntMap.IntMap String)
@@ -124,13 +129,13 @@ sslClientConn _desc onCerts h = do
 #endif
     let tcp = defaultParams
             { pConnectVersion = TLS10
-            , pAllowedVersions = [ TLS10, TLS11 ]
+            , pAllowedVersions = [ TLS10, TLS11, TLS12 ]
             , pCiphers = ciphersuite_all
             , onCertificatesRecv = onCerts
             }
     gen <- makeSystem
     istate <- client tcp gen h
-    _ <- handshake istate
+    handshake istate
     return ConnInfo
         { connRead = recvD istate
         , connWrite = sendData istate . L.fromChunks . (:[])
@@ -144,15 +149,14 @@ sslClientConn _desc onCerts h = do
   where
     recvD istate = do
         x <- recvData istate
-        if L.null x
+        if S.null x
             then recvD istate
-            else return $ S.concat $ L.toChunks x
-            -- Although a 'concat' seems like a bad idea, at
-            -- least on tls-0.8.4 it's guaranteed to always
-            -- return a lazy bytestring with a single chunk.
+            else return x
 
-getSocket :: String -> Int -> IO NS.Socket
-getSocket host' port' = do
+getSocket :: String -> Int -> Maybe SocksConf -> IO NS.Socket
+getSocket host' port' (Just socksConf) = do
+    socksConnectWith socksConf host' (PortNumber $ fromIntegral port')
+getSocket host' port' Nothing = do
     let hints = NS.defaultHints {
                           NS.addrFlags = [NS.AI_ADDRCONFIG]
                         , NS.addrSocketType = NS.Stream
